@@ -23,7 +23,6 @@ import imp
 import os
 
 import six
-
 from ansible import constants
 from ansible.errors import AnsibleError
 
@@ -38,6 +37,7 @@ except ImportError:
 import yaml
 from yaml.composer import Composer
 from yaml.constructor import Constructor
+import ruamel.yaml
 
 try:
     from ansible.utils import parse_yaml_from_file
@@ -50,12 +50,20 @@ except ImportError:
     from ansible.parsing.dataloader import DataLoader
     from ansible.template import Templar
     from ansible.parsing.mod_args import ModuleArgsParser
-    from ansible.plugins import module_loader
+    from ansible.parsing.yaml.constructor import AnsibleConstructor
+    from ansible.parsing.yaml.loader import AnsibleLoader
     from ansible.errors import AnsibleParserError
     ANSIBLE_VERSION = 2
 
+    # ansible-lint doesn't need/want to know about encrypted secrets, but it needs
+    # Ansible 2.3+ allows encrypted secrets within yaml files, so we pass a string
+    # as the password to enable such yaml files to be opened and parsed successfully.
+    DEFAULT_VAULT_PASSWORD = 'x'
+
     def parse_yaml_from_file(filepath):
         dl = DataLoader()
+        if hasattr(dl, 'set_vault_password'):
+            dl.set_vault_password(DEFAULT_VAULT_PASSWORD)
         return dl.load_from_file(filepath)
 
     def path_dwim(basedir, given):
@@ -68,17 +76,22 @@ except ImportError:
         dl.set_basedir(basedir)
         templar = Templar(dl, variables=templatevars)
         return templar.template(varname, **kwargs)
+    try:
+        from ansible.plugins import module_loader
+    except ImportError:
+        from ansible.plugins.loader import module_loader
 
 LINE_NUMBER_KEY = '__line__'
 FILENAME_KEY = '__file__'
 
 VALID_KEYS = [
     'name', 'action', 'when', 'async', 'poll', 'notify',
-    'first_available_file', 'include', 'tags', 'register', 'ignore_errors',
-    'delegate_to', 'local_action', 'transport', 'remote_user', 'sudo', 'sudo_user',
-    'sudo_pass', 'when', 'connection', 'environment', 'args',
-    'any_errors_fatal', 'changed_when', 'failed_when', 'always_run', 'delay', 'retries', 'until',
-    'su', 'su_user', 'su_pass', 'no_log', 'run_once',
+    'first_available_file', 'include', 'include_tasks', 'import_tasks', 'import_playbook',
+    'tags', 'register', 'ignore_errors', 'delegate_to',
+    'local_action', 'transport', 'remote_user', 'sudo',
+    'sudo_user', 'sudo_pass', 'when', 'connection', 'environment', 'args', 'always_run',
+    'any_errors_fatal', 'changed_when', 'failed_when', 'check_mode', 'delay',
+    'retries', 'until', 'su', 'su_user', 'su_pass', 'no_log', 'run_once',
     'become', 'become_user', 'become_method', FILENAME_KEY,
 ]
 
@@ -190,9 +203,12 @@ def play_children(basedir, item, parent_type, playbook_dir):
         'post_tasks': _taskshandlers_children,
         'block': _taskshandlers_children,
         'include': _include_children,
+        'import_playbook': _include_children,
         'roles': _roles_children,
         'dependencies': _roles_children,
         'handlers': _taskshandlers_children,
+        'include_tasks': _include_children,
+        'import_tasks': _include_children,
     }
     (k, v) = item
     play_library = os.path.join(os.path.abspath(basedir), 'library')
@@ -222,38 +238,57 @@ def _taskshandlers_children(basedir, k, v, parent_type):
     results = []
     for th in v:
         if 'include' in th:
-            # taskshandlers_children only get's called for playbooks, thus the
-            # actual type of the included tasks is the section containing the
-            # include, e.g. tasks, pre_tasks, or handlers.
-            assert(parent_type == 'playbook')
-            playbook_section = k
-            results.append({
-                'path': path_dwim(basedir, th['include']),
-                'type': playbook_section
-            })
+            append_children(th['include'], basedir, k, parent_type, results)
+        elif 'include_tasks' in th:
+            append_children(th['include_tasks'], basedir, k, parent_type, results)
+        elif 'import_playbook' in th:
+            append_children(th['import_playbook'], basedir, k, parent_type, results)
+        elif 'import_tasks' in th:
+            append_children(th['import_tasks'], basedir, k, parent_type, results)
+        elif 'import_role' in th:
+            results.extend(_roles_children(basedir, k, [th['import_role'].get('name')], parent_type,
+                                           main=th['import_role'].get('tasks_from', 'main')))
+        elif 'include_role' in th:
+            results.extend(_roles_children(basedir, k, [th['include_role'].get('name')],
+                                           parent_type,
+                                           main=th['include_role'].get('tasks_from', 'main')))
         elif 'block' in th:
             results.extend(_taskshandlers_children(basedir, k, th['block'], parent_type))
             if 'rescue' in th:
                 results.extend(_taskshandlers_children(basedir, k, th['rescue'], parent_type))
             if 'always' in th:
                 results.extend(_taskshandlers_children(basedir, k, th['always'], parent_type))
-
     return results
 
 
-def _roles_children(basedir, k, v, parent_type):
+def append_children(taskhandler, basedir, k, parent_type, results):
+    # when taskshandlers_children is called for playbooks, the
+    # actual type of the included tasks is the section containing the
+    # include, e.g. tasks, pre_tasks, or handlers.
+    if parent_type == 'playbook':
+        playbook_section = k
+    else:
+        playbook_section = parent_type
+    results.append({
+        'path': path_dwim(basedir, taskhandler),
+        'type': playbook_section
+    })
+
+
+def _roles_children(basedir, k, v, parent_type, main='main'):
     results = []
     for role in v:
         if isinstance(role, dict):
             if 'role' in role or 'name' in role:
                 if 'tags' not in role or 'skip_ansible_lint' not in role['tags']:
                     results.extend(_look_for_role_files(basedir,
-                                                        role.get('role', role.get('name'))))
-            else:
+                                                        role.get('role', role.get('name')),
+                                                        main=main))
+            elif k != 'dependencies':
                 raise SystemExit('role dict {0} does not contain a "role" '
                                  'or "name" key'.format(role))
         else:
-            results.extend(_look_for_role_files(basedir, role))
+            results.extend(_look_for_role_files(basedir, role, main=main))
     return results
 
 
@@ -273,7 +308,7 @@ def _rolepath(basedir, role):
         path_dwim(
             basedir, os.path.join('..', '..', '..', 'roles', role)
         ),
-        path_dwim(basedir, os.path.join('..', '..', role))
+        path_dwim(basedir, os.path.join('..', '..', role)),
     ]
 
     if constants.DEFAULT_ROLES_PATH:
@@ -284,6 +319,8 @@ def _rolepath(basedir, role):
             loc = os.path.expanduser(loc)
             possible_paths.append(path_dwim(loc, role))
 
+    possible_paths.append(path_dwim(basedir, ''))
+
     for path_option in possible_paths:
         if os.path.isdir(path_option):
             role_path = path_option
@@ -295,7 +332,7 @@ def _rolepath(basedir, role):
     return role_path
 
 
-def _look_for_role_files(basedir, role):
+def _look_for_role_files(basedir, role, main='main'):
     role_path = _rolepath(basedir, role)
     if not role_path:
         return []
@@ -303,11 +340,14 @@ def _look_for_role_files(basedir, role):
     results = []
 
     for th in ['tasks', 'handlers', 'meta']:
-        for ext in ('.yml', '.yaml'):
-            thpath = os.path.join(role_path, th, 'main' + ext)
-            if os.path.exists(thpath):
-                results.append({'path': thpath, 'type': th})
-                break
+        current_path = os.path.join(role_path, th)
+        for dir, subdirs, files in os.walk(current_path):
+            for file in files:
+                file_ignorecase = file.lower()
+                if file_ignorecase.endswith(('.yml', '.yaml')):
+                    thpath = os.path.join(dir, file)
+                    results.append({'path': thpath, 'type': th})
+
     return results
 
 
@@ -363,10 +403,15 @@ def normalize_task_v2(task):
     result['action'] = dict(__ansible_module__=action)
 
     if '_raw_params' in arguments:
-        result['action']['__ansible_arguments__'] = arguments['_raw_params'].split()
+        result['action']['__ansible_arguments__'] = arguments['_raw_params'].split(' ')
         del(arguments['_raw_params'])
     else:
         result['action']['__ansible_arguments__'] = list()
+
+    if 'argv' in arguments and not result['action']['__ansible_arguments__']:
+        result['action']['__ansible_arguments__'] = arguments['argv']
+        del(arguments['argv'])
+
     result['action'].update(arguments)
     return result
 
@@ -399,7 +444,7 @@ def normalize_task_v1(task):
                     else:
                         # Tasks that include playbooks (rather than task files)
                         # can get here
-                        # https://github.com/willthames/ansible-lint/issues/138
+                        # https://github.com/ansible/ansible-lint/issues/138
                         raise RuntimeError("Was not expecting value %s of type %s for key %s\n"
                                            "Task: %s. Check the syntax of your playbook using "
                                            "ansible-playbook --syntax-check" %
@@ -437,10 +482,10 @@ def task_to_str(task):
     if name:
         return name
     action = task.get("action")
-    args = " " .join(["{0}={1}".format(k, v) for (k, v) in action.items()
+    args = " ".join([u"{0}={1}".format(k, v) for (k, v) in action.items()
                      if k not in ["__ansible_module__", "__ansible_arguments__"]] +
-                     action.get("__ansible_arguments__"))
-    return "{0} {1}".format(action["__ansible_module__"], args)
+                    action.get("__ansible_arguments__"))
+    return u"{0} {1}".format(action["__ansible_module__"], args)
 
 
 def extract_from_list(blocks, candidates):
@@ -478,12 +523,24 @@ def get_action_tasks(yaml, file):
     block_rescue_always = ('block', 'rescue', 'always')
     tasks[:] = [task for task in tasks if all(k not in task for k in block_rescue_always)]
 
-    return [task for task in tasks if 'include' not in task.keys()]
+    return [task for task in tasks if
+            set(['include', 'include_tasks',
+                'import_playbook', 'import_tasks']).isdisjoint(task.keys())]
 
 
 def get_normalized_tasks(yaml, file):
     tasks = get_action_tasks(yaml, file)
-    return [normalize_task(task, file['path']) for task in tasks]
+    res = []
+    for task in tasks:
+        # An empty `tags` block causes `None` to be returned if
+        # the `or []` is not present - `task.get('tags', [])`
+        # does not suffice.
+        if 'skip_ansible_lint' in (task.get('tags') or []):
+            # No need to normalize_task is we are skipping it.
+            continue
+        res.append(normalize_task(task, file['path']))
+
+    return res
 
 
 def parse_yaml_linenumbers(data, filename):
@@ -500,16 +557,111 @@ def parse_yaml_linenumbers(data, filename):
         return node
 
     def construct_mapping(node, deep=False):
-        mapping = Constructor.construct_mapping(loader, node, deep=deep)
-        mapping[LINE_NUMBER_KEY] = node.__line__
+        if ANSIBLE_VERSION < 2:
+            mapping = Constructor.construct_mapping(loader, node, deep=deep)
+        else:
+            mapping = AnsibleConstructor.construct_mapping(loader, node, deep=deep)
+        if hasattr(node, '__line__'):
+            mapping[LINE_NUMBER_KEY] = node.__line__
+        else:
+            mapping[LINE_NUMBER_KEY] = mapping._line_number
         mapping[FILENAME_KEY] = filename
         return mapping
 
     try:
-        loader = yaml.Loader(data)
+        if ANSIBLE_VERSION < 2:
+            loader = yaml.Loader(data)
+        else:
+            import inspect
+            kwargs = {}
+            if 'vault_password' in inspect.getargspec(AnsibleLoader.__init__).args:
+                kwargs['vault_password'] = DEFAULT_VAULT_PASSWORD
+            loader = AnsibleLoader(data, **kwargs)
         loader.compose_node = compose_node
         loader.construct_mapping = construct_mapping
         data = loader.get_single_data()
     except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
         raise SystemExit("Failed to parse YAML in %s: %s" % (filename, str(e)))
     return data
+
+
+def get_first_cmd_arg(task):
+    try:
+        if 'cmd' in task['action']:
+            first_cmd_arg = task['action']['cmd'].split()[0]
+        else:
+            first_cmd_arg = task['action']['__ansible_arguments__'][0]
+    except IndexError:
+        return None
+    return first_cmd_arg
+
+
+def append_skipped_rules(pyyaml_data, file_text, file_type):
+    """ Uses ruamel.yaml to parse comments then adds a
+        skipped_rules list to the task (or meta yaml block)
+    """
+    yaml = ruamel.yaml.YAML()
+    ruamel_data = yaml.load(file_text)
+
+    if file_type in ('tasks', 'handlers'):
+        ruamel_tasks = ruamel_data
+        pyyaml_tasks = pyyaml_data
+    elif file_type == 'playbook':
+        try:
+            ruamel_tasks = []
+            pyyaml_tasks = []
+            for ruamel_play, pyyaml_play in zip(ruamel_data, pyyaml_data):
+                ruamel_tasks.extend(ruamel_play.get('tasks'))
+                pyyaml_tasks.extend(pyyaml_play.get('tasks'))
+        except (AttributeError, TypeError):
+            return pyyaml_data
+    elif file_type == 'meta':
+        if not isinstance(pyyaml_data, list):
+            return pyyaml_data
+        ruamel_tasks = [ruamel_data]
+        pyyaml_tasks = pyyaml_data
+    else:
+        return pyyaml_data
+
+    if len(ruamel_tasks) != len(pyyaml_tasks):
+        return pyyaml_data
+
+    for ruamel_task, pyyaml_task in zip(ruamel_tasks, pyyaml_tasks):
+        skipped_rules = _get_rule_skips_from_task(ruamel_task)
+        if skipped_rules:
+            pyyaml_task['skipped_rules'] = skipped_rules
+
+    return pyyaml_data
+
+
+def _get_rule_skips_from_task(task):
+    def traverse_yaml(obj):
+        task_comment_obj_strs.append(str(obj.ca.items))
+        if isinstance(obj, dict):
+            for key, val in obj.items():
+                if isinstance(val, (dict, list)):
+                    traverse_yaml(val)
+        elif isinstance(obj, list):
+            for e in obj:
+                if isinstance(e, (dict, list)):
+                    traverse_yaml(e)
+        else:
+            return
+
+    task_comment_obj_strs = []
+    traverse_yaml(task)
+
+    rule_id_list = []
+    for comment_obj_str in task_comment_obj_strs:
+        for line in comment_obj_str.split('\\n'):
+            rule_id_list.extend(get_rule_skips_from_line(line))
+
+    return rule_id_list
+
+
+def get_rule_skips_from_line(line):
+    rule_id_list = []
+    if '# noqa' in line:
+        noqa_text = line.split('# noqa')[1]
+        rule_id_list = noqa_text.split()
+    return rule_id_list
